"""ffmpeg_frame_writer - Module that implements frame writing using ffmpeg"""
from pathlib import Path
import os
from tqdm import trange
import ffmpeg

from .frame_writer import FrameWriter

class FFmpegFrameWriter(FrameWriter):
    """FFmpegFrameWriter implementation that writes to disk a video using ffmpeg process"""
    def __init__(self):
        super().__init__()
        self.write_process = None

    def write(self, video: "VREVideo", out_path: Path, start_frame: int = 0, end_frame: int | None = None):
        out_path = Path(out_path)
        assert self.write_process is None, self.write_process
        assert out_path.suffix == ".mp4", out_path
        assert isinstance(start_frame, int) and start_frame >= 0, start_frame

        self.write_process = (
            ffmpeg
            .input("pipe:0", format="rawvideo", pix_fmt="rgb24", s=f"{video.shape[2]}x{video.shape[1]}", r=video.fps)
            .output(str(out_path), pix_fmt="yuv420p", vcodec="libx264")
            .overwrite_output()
            .run_async(pipe_stdin=True, pipe_stderr=-3, pipe_stdout=-3) # -3 = subprocess.DEVNULL
        )

        assert not out_path.exists(), out_path
        out_path.parent.mkdir(exist_ok=True, parents=True)
        start_frame = start_frame or 0
        end_frame = end_frame or len(video)
        assert start_frame >= 0 and end_frame <= len(video), (start_frame, end_frame)

        try:
            for i in trange(start_frame, end_frame, disable=os.getenv("VRE_VIDEO_PBAR", "1") == "0",
                            desc=f"[FFmpegFrameWriter] {out_path}"):
                self.write_process.stdin.write(video[i].tobytes())
        finally:
            self.write_process.stdin.close()
            self.write_process.wait()
            self.write_process = None



"""pil_frame_reader - module for PIL-based functionality"""
from pathlib import Path
from natsort import natsorted
import numpy as np
from .frame_reader import FrameReader
from ..utils import image_read

class PILFrameReader(FrameReader):
    """implements FrameReader using Pillow to read frame by frame"""
    def __init__(self, path: str | Path, fps: int = 1):
        self.path = Path(path)
        self.frame_paths = natsorted(list(self.path.iterdir()), key=lambda p: p.name)
        self.data: list[np.ndarray] = [image_read(x) for x in self.frame_paths]
        _shp = self.data[0].shape
        assert all(x.shape == _shp for x in self.data), f"Not all shapes of all images are equal to first image: {_shp}"
        assert all(x.dtype == np.uint8 for x in self.data), f"Not all data is uint8: {set(x.dtype for x in self.data)}"
        self.data = np.array(self.data)
        self._fps = fps

    @property
    def shape(self) -> tuple[int, int, int, int]:
        return (len(self.data), *self.data[0].shape)

    @property
    def fps(self):
        return self._fps

    def get_one_frame(self, ix):
        return self.data[ix]

    def __getitem__(self, ix: int | list[int] | np.ndarray | slice) -> np.ndarray:
        return self.data[ix]
